'''
A pure python version of the hdr_histogram code

Ported from
https://github.com/HdrHistogram/HdrHistogram (Java)
https://github.com/HdrHistogram/HdrHistogram_c (C)

Written by Alec Hothan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
import math
import sys
from hdrh.iterators import AllValuesIterator
from hdrh.iterators import RecordedIterator
from hdrh.iterators import PercentileIterator
from hdrh.iterators import LinearIterator
from hdrh.iterators import LogIterator
from hdrh.codec import HdrHistogramEncoder

def get_bucket_count(value, subb_count, unit_mag):
    smallest_untrackable_value = subb_count << unit_mag
    buckets_needed = 1
    while smallest_untrackable_value <= value:
        if smallest_untrackable_value > sys.maxint / 2:
            return buckets_needed + 1
        smallest_untrackable_value <<= 1
        buckets_needed += 1
    return buckets_needed

class HdrHistogram(object):
    '''This class supports the recording and analyzing of sampled data value
    counts across a configurable integer value range with configurable value
    precision within the range. Value precision is expressed as the number of
    significant digits in the value recording, and provides control over value
    quantization behavior across the value range and the subsequent value
    resolution at any given level.

    For example, a Histogram could be configured to track the counts of
    observed integer values between 0 and 3,600,000,000 while maintaining a
    value precision of 3 significant digits across that range. Value
    quantization within the range will thus be no larger than 1/1,000th
    (or 0.1%) of any value. This example Histogram could be used to track and
    analyze the counts of observed response times ranging between 1 microsecond
    and 1 hour in magnitude, while maintaining a value resolution of 1
    microsecond up to 1 millisecond, a resolution of 1 millisecond (or better)
    up to one second, and a resolution of 1 second (or better) up to 1,000
    seconds. At it's maximum tracked value (1 hour), it would still maintain a
    resolution of 3.6 seconds (or better).
    '''

    def __init__(self,
                 lowest_trackable_value,
                 highest_trackable_value,
                 significant_figures,
                 word_size=8,
                 b64_wrap=True,
                 hdr_payload=None):
        '''Create a new histogram with given arguments

        Params:
            lowest_trackable_value The lowest value that can be discerned
                (distinguished from 0) by the histogram.
                Must be a positive integer that is >= 1.
                May be internally rounded down to nearest power of 2.
            highest_trackable_value The highest value to be tracked by the
                histogram. Must be a positive integer that is >=
                (2 * lowest_trackable_value).
            significant_figures The number of significant decimal digits to
                which the histogram will maintain value resolution and
                separation. Must be a non-negative integer between 0 and 5.
            word_size size of counters in bytes, only 2, 4, 8-byte counters
                are supported (default is 8-byte or 64-bit counters)
            b64_wrap specifies if the encoding of this histogram should use
                base64 wrapping (only useful if you need to encode the histogram
                to save somewhere or send over the wire. By default base64
                encoding is assumed
            hdr_payload only used for associating an existing payload created
                from decoding an encoded histograme
        Exceptions:
            ValueError if the word_size value is unsupported
                if significant_figures is invalid
        '''
        if significant_figures < 1 or significant_figures > 5:
            raise ValueError('Invalid significant_figures')
        self.lowest_trackable_value = lowest_trackable_value
        self.highest_trackable_value = highest_trackable_value
        self.significant_figures = significant_figures
        self.unit_magnitude = int(math.floor(math.log(lowest_trackable_value) /
                                             math.log(2)))
        largest_value_single_unit_res = 2 * math.pow(10, significant_figures)
        subb_count_mag = int(math.ceil(math.log(largest_value_single_unit_res) /
                                       math.log(2)))
        self.sub_bucket_half_count_magnitude = subb_count_mag - 1 if subb_count_mag > 1 else 0
        self.sub_bucket_count = int(math.pow(2, self.sub_bucket_half_count_magnitude + 1))
        self.sub_bucket_half_count = self.sub_bucket_count / 2
        self.sub_bucket_mask = (self.sub_bucket_count - 1) << self.unit_magnitude
        self.bucket_count = get_bucket_count(highest_trackable_value,
                                             self.sub_bucket_count,
                                             self.unit_magnitude)
        self.min_value = sys.maxint
        self.max_value = 0
        self.total_count = 0
        self.counts_len = (self.bucket_count + 1) * (self.sub_bucket_count / 2)
        self.word_size = word_size

        if hdr_payload:
            payload = hdr_payload.payload
            self.int_to_double_conversion_ratio = payload.conversion_ratio_bits
            results = hdr_payload.init_counts(self.counts_len)
            if results['total']:
                self.set_internal_tacking_values(results['min_nonzero_index'],
                                                 results['max_nonzero_index'],
                                                 results['total'])
        else:
            self.int_to_double_conversion_ratio = 1.0

        # to encode this histogram into a compressed/base64 format ready
        # to be exported
        self.b64_wrap = b64_wrap
        self.encoder = HdrHistogramEncoder(self, b64_wrap, hdr_payload)
        # the counters reside directly in the payload object
        # allocated by the encoder
        # so that compression for wire transfer can be done without copy
        self.counts = self.encoder.get_counts()
        self.start_time_stamp_msec = 0
        self.end_time_stamp_msec = 0

    def _clz(self, value):
        """calculate the leading zeros, equivalent to C __builtin_clzll()
        value in hex:
        value = 1 clz = 63
        value = 2 clz = 62
        value = 4 clz = 61
        value = 1000 clz = 51
        value = 1000000 clz = 39
        """
        return 63 - (len(bin(value)) - 3)

    def _get_bucket_index(self, value):
        # smallest power of 2 containing value
        pow2ceiling = 64 - self._clz(int(value) | self.sub_bucket_mask)
        return int(pow2ceiling - self.unit_magnitude -
                   (self.sub_bucket_half_count_magnitude + 1))

    def _get_sub_bucket_index(self, value, bucket_index):
        return int(value) >> (bucket_index + self.unit_magnitude)

    def _counts_index(self, bucket_index, sub_bucket_index):
        # Calculate the index for the first entry in the bucket:
        # (The following is the equivalent of ((bucket_index + 1) * subBucketHalfCount) ):
        bucket_base_index = (bucket_index + 1) << self.sub_bucket_half_count_magnitude
        # Calculate the offset in the bucket:
        offset_in_bucket = sub_bucket_index - self.sub_bucket_half_count
        # The following is the equivalent of
        # ((sub_bucket_index  - subBucketHalfCount) + bucketBaseIndex
        return bucket_base_index + offset_in_bucket

    def _counts_index_for(self, value):
        bucket_index = self._get_bucket_index(value)
        sub_bucket_index = self._get_sub_bucket_index(value, bucket_index)
        return self._counts_index(bucket_index, sub_bucket_index)

    def record_value(self, value, count=1):
        '''Record a new value into the histogram

        Args:
            value: the value to record (must be in the valid range)
            count: incremental count (defaults to 1)
        '''
        if value < 0:
            return False
        counts_index = self._counts_index_for(value)
        if (counts_index < 0) or (self.counts_len <= counts_index):
            return False
        self.counts[counts_index] += count
        self.total_count += count
        self.min_value = min(self.min_value, value)
        self.max_value = max(self.max_value, value)
        return True

    def record_corrected_value(self, value, expected_interval, count=1):
        '''Record a new value into the histogram and correct for
        coordinated omission if needed

        Args:
            value: the value to record (must be in the valid range)
            expected_interval: the expected interval between 2 value samples
            count: incremental count (defaults to 1)
        '''
        while True:
            if not self.record_value(value, count):
                return False
            if value <= expected_interval or expected_interval <= 0:
                return True
            value -= expected_interval

    def get_count_at_index(self, index):
        if index >= self.counts_len:
            raise IndexError()
        # some decoded (read-only) histograms may have truncated
        # counts arrays, we return zero for any index that is passed the array
        if index >= self.encoder.payload.counts_len:
            return 0
        return self.counts[index]

    def get_count_at_sub_bucket(self, bucket_index, sub_bucket_index):
        # Calculate the index for the first entry in the bucket:
        # (The following is the equivalent of ((bucket_index + 1) * subBucketHalfCount) )
        bucket_base_index = (bucket_index + 1) << self.sub_bucket_half_count_magnitude
        # Calculate the offset in the bucket:
        offset_in_bucket = sub_bucket_index - self.sub_bucket_half_count
        # The following is the equivalent of
        # (sub_bucket_index - subBucketHalfCount) + bucketBaseIndex
        counts_index = bucket_base_index + offset_in_bucket
        return self.counts[counts_index]

    def get_value_from_sub_bucket(self, bucket_index, sub_bucket_index):
        return sub_bucket_index << (bucket_index + self.unit_magnitude)

    def get_value_from_index(self, index):
        bucket_index = (index >> self.sub_bucket_half_count_magnitude) - 1
        sub_bucket_index = (index & (self.sub_bucket_half_count - 1)) + \
            self.sub_bucket_half_count
        if bucket_index < 0:
            sub_bucket_index -= self.sub_bucket_half_count
            bucket_index = 0
        return self.get_value_from_sub_bucket(bucket_index, sub_bucket_index)

    def get_lowest_equivalent_value(self, value):
        bucket_index = self._get_bucket_index(value)
        sub_bucket_index = self._get_sub_bucket_index(value, bucket_index)

        lowest_equivalent_value = self.get_value_from_sub_bucket(bucket_index,
                                                                 sub_bucket_index)
        return lowest_equivalent_value

    def get_highest_equivalent_value(self, value):
        bucket_index = self._get_bucket_index(value)
        sub_bucket_index = self._get_sub_bucket_index(value, bucket_index)

        lowest_equivalent_value = self.get_value_from_sub_bucket(bucket_index,
                                                                 sub_bucket_index)
        if sub_bucket_index >= self.sub_bucket_count:
            bucket_index += 1
        size_of_equivalent_value_range = 1 << (self.unit_magnitude + bucket_index)
        next_non_equivalent_value = lowest_equivalent_value + size_of_equivalent_value_range

        return next_non_equivalent_value - 1

    def get_target_count_at_percentile(self, percentile):
        requested_percentile = min(percentile, 100.0)
        count_at_percentile = int(((requested_percentile * self.total_count / 100)) + 0.5)
        return max(count_at_percentile, 1)

    def get_value_at_percentile(self, percentile):
        '''Get the value for a given percentile

        Args:
            percentile: a float in [0.0..100.0]
        Returns:
            the value for the given percentile
        '''
        count_at_percentile = self.get_target_count_at_percentile(percentile)
        total = 0
        for index in xrange(self.counts_len):
            total += self.get_count_at_index(index)
            if total >= count_at_percentile:
                value_at_index = self.get_value_from_index(index)
                if percentile:
                    return self.get_highest_equivalent_value(value_at_index)
                return self.get_lowest_equivalent_value(value_at_index)
        return 0

    def get_percentile_to_value_dict(self, percentile_list):
        '''A faster alternative to query values for a list of percentiles.

        Args:
            percentile_list: a list of percentiles in any order, dups will be ignored
            each element in the list must be a float value in [0.0 .. 100.0]
        Returns:
            a dict of percentile values indexed by the percentile
        '''
        result = {}
        total = 0
        percentile_list_index = 0
        count_at_percentile = 0
        # remove dups and sort
        percentile_list = list(set(percentile_list))
        percentile_list.sort()

        for index in xrange(self.counts_len):
            total += self.get_count_at_index(index)
            while True:
                # recalculate target based on next requested percentile
                if not count_at_percentile:
                    if percentile_list_index == len(percentile_list):
                        return result
                    percentile = percentile_list[percentile_list_index]
                    percentile_list_index += 1
                    if percentile > 100:
                        return result
                    count_at_percentile = self.get_target_count_at_percentile(percentile)

                if total >= count_at_percentile:
                    value_at_index = self.get_value_from_index(index)
                    if percentile:
                        result[percentile] = self.get_highest_equivalent_value(value_at_index)
                    else:
                        result[percentile] = self.get_lowest_equivalent_value(value_at_index)
                    count_at_percentile = 0
                else:
                    break
        return result

    def get_total_count(self):
        return self.total_count

    def get_count_at_value(self, value):
        counts_index = self._counts_index_for(value)
        return self.counts[counts_index]

    def values_are_equivalent(self, val1, val2):
        '''Check whether 2 values are equivalent (meaning they
        are in the same bucket/range)

        Returns:
            true if the 2 values are equivalent
        '''
        return self.get_lowest_equivalent_value(val1) == self.get_lowest_equivalent_value(val2)

    def get_max_value(self):
        if 0 == self.max_value:
            return 0
        return self.get_highest_equivalent_value(self.max_value)

    def get_min_value(self):
        if 0 < self.counts[0] or self.total_count == 0:
            return 0
        if sys.maxint == self.min_value:
            return sys.maxint
        return self.get_lowest_equivalent_value(self.min_value)

    def _hdr_size_of_equiv_value_range(self, value):
        bucket_index = self._get_bucket_index(value)
        sub_bucket_index = self._get_sub_bucket_index(value, bucket_index)
        if sub_bucket_index >= self.sub_bucket_count:
            bucket_index += 1
        return 1 << (self.unit_magnitude + bucket_index)

    def _hdr_median_equiv_value(self, value):
        return self.get_lowest_equivalent_value(value) + \
            (self._hdr_size_of_equiv_value_range(value) >> 1)

    def get_mean_value(self):
        if not self.total_count:
            return 0.0
        total = 0
        itr = self.get_recorded_iterator()
        for item in itr:
            total += itr.count_at_this_value * self._hdr_median_equiv_value(item.value_iterated_to)
        return float(total) / self.total_count

    def get_stddev(self):
        if not self.total_count:
            return 0.0
        mean = self.get_mean_value()
        geometric_dev_total = 0.0
        for item in self.get_recorded_iterator():
            dev = (self._hdr_median_equiv_value(item.value_iterated_to) * 1.0) - mean
            geometric_dev_total += (dev * dev) * item.count_added_in_this_iter_step
        return math.sqrt(geometric_dev_total / self.total_count)

    def reset(self):
        '''Reset the histogram to a pristine state
        '''
        for index in xrange(self.counts_len):
            self.counts[index] = 0
        self.total_count = 0
        self.min_value = sys.maxint
        self.max_value = 0

    def __iter__(self):
        '''Returns the recorded iterator if iter(self) is called
        '''
        return RecordedIterator(self)

    def get_all_values_iterator(self):
        return AllValuesIterator(self)

    def get_recorded_iterator(self):
        return RecordedIterator(self)

    def get_percentile_iterator(self, ticks_per_half_distance):
        return PercentileIterator(self, ticks_per_half_distance)

    def get_linear_iterator(self, value_units_per_bucket):
        return LinearIterator(self, value_units_per_bucket)

    def get_log_iterator(self, value_units_first_bucket, log_base):
        return LogIterator(self, value_units_first_bucket, log_base)

    def encode(self):
        '''Encode this histogram
        Return:
            a string containing the base64 encoded compressed histogram (V1 format)
        '''
        return self.encoder.encode()

    def adjust_internal_tacking_values(self,
                                       min_non_zero_index,
                                       max_index,
                                       total_added):
        '''Called during decoding and add to adjust the new min/max value and
        total count

        Args:
            min_non_zero_index min nonzero index of all added counts (-1 if none)
            max_index max index of all added counts (-1 if none)
        '''
        if max_index >= 0:
            max_value = self.get_highest_equivalent_value(self.get_value_from_index(max_index))
            self.max_value = max(self.max_value, max_value)
        if min_non_zero_index >= 0:
            min_value = self.get_value_from_index(min_non_zero_index)
            self.min_value = min(self.min_value, min_value)
        self.total_count += total_added

    def set_internal_tacking_values(self,
                                    min_non_zero_index,
                                    max_index,
                                    total_added):
        '''Called during decoding and add to adjust the new min/max value and
        total count

        Args:
            min_non_zero_index min nonzero index of all added counts (-1 if none)
            max_index max index of all added counts (-1 if none)
        '''
        if max_index >= 0:
            self.max_value = self.get_highest_equivalent_value(self.get_value_from_index(max_index))
        if min_non_zero_index >= 0:
            self.min_value = self.get_value_from_index(min_non_zero_index)
        self.total_count = total_added

    def get_counts_array_index(self, value):
        '''Return the index in the counts array for a given value
        '''
        if value < 0:
            raise ValueError("Histogram recorded value cannot be negative.")

        bucket_index = self._get_bucket_index(value)
        sub_bucket_index = self._get_sub_bucket_index(value, bucket_index)
        # Calculate the index for the first entry in the bucket:
        bucket_base_index = (bucket_index + 1) << self.sub_bucket_half_count_magnitude
        # The following is the equivalent of ((bucket_index + 1) * sub_bucket_half_count)
        # Calculate the offset in the bucket (can be negative for first bucket):
        offset_in_bucket = sub_bucket_index - self.sub_bucket_half_count
        # The following is the equivalent of
        # ((sub_bucket_index  - sub_bucket_half_count) + bucket_base_index
        return bucket_base_index + offset_in_bucket

    def get_start_time_stamp(self):
        return self.start_time_stamp_msec

    def set_start_time_stamp(self, time_stamp_msec):
        '''Set the start time stamp value associated with this histogram to a given value.
        Params:
            time_stamp_msec the value to set the time stamp to,
                [by convention] in msec since the epoch.
        '''
        self.start_time_stamp_msec = time_stamp_msec

    def get_end_time_stamp(self):
        return self.end_time_stamp_msec

    def set_end_time_stamp(self, time_stamp_msec):
        '''Set the end time stamp value associated with this histogram to a given value.
        Params:
            time_stamp_msec the value to set the time stamp to,
                [by convention] in msec since the epoch.
        '''
        self.end_time_stamp_msec = time_stamp_msec

    def add(self, other_hist):
        highest_recordable_value = \
            self.get_highest_equivalent_value(self.get_value_from_index(self.counts_len - 1))
        if highest_recordable_value < other_hist.get_max_value():
            raise IndexError("The other histogram includes values that do not fit %d < %d" %
                             (highest_recordable_value, other_hist.get_max_value()))

        if (self.bucket_count == other_hist.bucket_count) and \
           (self.sub_bucket_count == other_hist.sub_bucket_count) and \
           (self.unit_magnitude == other_hist.unit_magnitude) and \
           (self.word_size == other_hist.word_size):

            # do an in-place addition of one array to another
            self.encoder.add(other_hist.encoder)

            self.total_count += other_hist.get_total_count()
            self.max_value = max(self.max_value, other_hist.get_max_value())
            self.min_value = min(self.get_min_value(), other_hist.get_min_value())
        else:
            # Arrays are not a direct match, so we can't just stream through and add them.
            # Instead, go through the array and add each non-zero value found at it's proper value:
            for index in xrange(other_hist.counts_len):
                other_count = other_hist.get_count_at_index(index)
                if other_count > 0:
                    self.record_value(other_hist.get_value_from_index(index), other_count)

        self.start_time_stamp_msec = \
            min(self.start_time_stamp_msec, other_hist.start_time_stamp_msec)
        self.end_time_stamp_msec = \
            max(self.end_time_stamp_msec, other_hist.end_time_stamp_msec)

    def decode_and_add(self, encoded_histogram):
        '''Decode an encoded histogram and add it to this histogram
        Args:
            encoded_histogram (string) an encoded histogram
                following the V1 format, such as one returned by the encode() method
        Exception:
            TypeError in case of base64 decode error
            HdrCookieException:
                the main header has an invalid cookie
                the compressed payload header has an invalid cookie
            HdrLengthException:
                the decompressed size is too small for the HdrPayload structure
                or is not aligned or is too large for the passed payload class
            zlib.error:
                in case of zlib decompression error
        '''
        other_hist = HdrHistogram.decode(encoded_histogram, self.b64_wrap)
        self.add(other_hist)

    @staticmethod
    def decode(encoded_histogram, b64_wrap=True):
        '''Decode an encoded histogram and return a new histogram instance that
        has been initialized with the decoded content
        Return:
            a new histogram instance representing the decoded content
        Exception:
            TypeError in case of base64 decode error
            HdrCookieException:
                the main header has an invalid cookie
                the compressed payload header has an invalid cookie
            HdrLengthException:
                the decompressed size is too small for the HdrPayload structure
                or is not aligned or is too large for the passed payload class
            zlib.error:
                in case of zlib decompression error
        '''
        hdr_payload = HdrHistogramEncoder.decode(encoded_histogram, b64_wrap)
        payload = hdr_payload.payload
        histogram = HdrHistogram(payload.lowest_trackable_value,
                                 payload.highest_trackable_value,
                                 payload.significant_figures,
                                 hdr_payload=hdr_payload)
        return histogram

    def get_word_size(self):
        return self.word_size



'''
A pure python version of the hdr_histogram code

Ported from
https://github.com/HdrHistogram/HdrHistogram (Java)
https://github.com/HdrHistogram/HdrHistogram_c (C)

Coding / Decoding of histograms using the
HdrHistogram V1 compressed transportable format

Written by Alec Hothan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import base64
import ctypes
from ctypes import BigEndianStructure
from ctypes import addressof
from ctypes import c_byte
from ctypes import c_ushort
from ctypes import c_uint
from ctypes import c_ulonglong

from pyhdrh import add_array
from pyhdrh import decode
from pyhdrh import encode
import zlib

V2_ENCODING_COOKIE_BASE = 0x1c849303
V2_COMPRESSION_COOKIE_BASE = 0x1c849304

# LEB128 + ZigZag require up to 9 bytes per word
V2_MAX_WORD_SIZE_IN_BYTES = 9

# allow at most 4 MB counter array size
# that represents 500,000 8 byte counters
MAX_COUNTS_SIZE = 4 * 1024 * 1024

def get_cookie_base(cookie):
    return cookie & ~0xf0

def get_word_size_in_bytes_from_cookie(cookie):
    if (get_cookie_base(cookie) == V2_ENCODING_COOKIE_BASE) or \
       (get_cookie_base(cookie) == V2_COMPRESSION_COOKIE_BASE):
        return V2_MAX_WORD_SIZE_IN_BYTES
    return (cookie & 0xf0) >> 4

def get_encoding_cookie():
    # LSBit of wordsize byte indicates TLZE Encoding
    return V2_ENCODING_COOKIE_BASE | 0x10

def get_compression_cookie():
    # LSBit of wordsize byte indicates TLZE Encoding
    return V2_COMPRESSION_COOKIE_BASE | 0x10

class HdrCookieException(Exception):
    pass

class HdrLengthException(Exception):
    pass

class HdrHistogramSettingsException(Exception):
    pass

# External encoding header
class ExternalHeader(BigEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("cookie", c_uint),
        ("length", c_uint)]
ext_header_size = ctypes.sizeof(ExternalHeader)

# Header for the zlib compressed part
class PayloadHeader(BigEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("cookie", c_uint),
        ("payload_len", c_uint),
        ("normalizing_index_offset", c_uint),
        ("significant_figures", c_uint),
        ("lowest_trackable_value", c_ulonglong),
        ("highest_trackable_value", c_ulonglong),
        ("conversion_ratio_bits", c_ulonglong)]
payload_header_size = ctypes.sizeof(PayloadHeader)

# list of supported payload counter ctypes, indexed by the word size
payload_counter_ctype = [None, None,
                         c_ushort,      # index 2
                         None,
                         c_uint,        # index 4
                         None, None, None,
                         c_ulonglong]   # index 8

class HdrPayload(object):
    '''A class that wraps the ctypes big endian struct that will hold the
    histogram wire format content (including the counters).
    '''
    def __init__(self, word_size, counts_len=0, compressed_payload=None):
        '''Two ways to use this class:
        - for an empty histogram (pass counts_len>0 and compressed_payload=None)
        - for a decoded histogram (counts_len=0 and compressed_payload!=None)

        Params:
            word_size counter size in bytes (2,4,8 byte counters are supported)
            counts_len number of counters to allocate
                ignored if a compressed payload is provided (not None)

            compressed_payload (string) a payload in zlib compressed form,
                decompress and decode the payload header.
                Caller must then invoke init_counts to pass in counts_len so that the
                counts array can be updated from the decoded varint buffer
                None if no compressed payload is to be associated to this instance
        '''
        self.word_size = word_size
        self.counts_len = counts_len
        self._data = None
        try:
            # ctype counter type
            self.counter_ctype = payload_counter_ctype[word_size]
        except IndexError:
            raise ValueError('Invalid word size')
        if not self.counter_ctype:
            raise ValueError('Invalid word size')
        if compressed_payload:
            self._decompress(compressed_payload)
        elif counts_len:
            self.payload = PayloadHeader()
            self.payload.cookie = get_encoding_cookie()
            self._init_counts()
        else:
            raise RuntimeError('counts_len cannot be zero')

    def _init_counts(self):
        self.counts = (self.counter_ctype * self.counts_len)()

    def init_counts(self, counts_len):
        '''Called after instantiating with a compressed payload
        Params:
            counts_len counts size to use based on decoded settings in the header
        '''
        assert(self._data and counts_len and self.counts_len == 0)
        self.counts_len = counts_len
        self._init_counts()

        results = decode(self._data, payload_header_size, addressof(self.counts),
                         counts_len, self.word_size)
        # no longer needed
        self._data = None
        return results

    def get_counts(self):
        return self.counts

    def _decompress(self, compressed_payload):
        '''Decompress a compressed payload into this payload wrapper.
        Note that the decompressed buffer is saved in self._data and the
        counts array is not yet allocated.

        Args:
            compressed_payload (string) a payload in zlib compressed form
        Exception:
            HdrCookieException:
                the compressed payload has an invalid cookie
            HdrLengthException:
                the decompressed size is too small for the HdrPayload structure
                or is not aligned or is too large for the passed payload class
            HdrHistogramSettingsException:
                mismatch in the significant figures, lowest and highest
                         trackable value
        '''
        # make sure this instance is pristine
        if self._data:
            raise RuntimeError('Cannot decompress to an instance with payload')
        # Here it is important to keep a reference to the decompressed
        # string so that it does not get garbage collected
        self._data = zlib.decompress(compressed_payload)
        len_data = len(self._data)

        counts_size = len_data - payload_header_size
        if payload_header_size > counts_size > MAX_COUNTS_SIZE:
            raise HdrLengthException('Invalid size:' + str(len_data))

        # copy the first bytes for the header
        self.payload = PayloadHeader.from_buffer_copy(self._data)

        cookie = self.payload.cookie
        if get_cookie_base(cookie) != V2_ENCODING_COOKIE_BASE:
            raise HdrCookieException('Invalid cookie: %x' % cookie)
        word_size = get_word_size_in_bytes_from_cookie(cookie)
        if word_size != V2_MAX_WORD_SIZE_IN_BYTES:
            raise HdrCookieException('Invalid V2 cookie: %x' % cookie)

    def compress(self, counts_limit):
        '''Compress this payload instance
        Args:
            counts_limit how many counters should be encoded
                           starting from index 0 (can be 0),
        Return:
            the compressed payload (python string)
        '''
        if self.payload:
            # worst case varint encoded length is when each counter is at the maximum value
            # in this case 1 more byte per counter is needed due to the more bits
            varint_len = counts_limit * (self.word_size + 1)
            # allocate enough space to fit the header and the varint string
            encode_buf = (c_byte * (payload_header_size + varint_len))()

            # encode past the payload header
            varint_len = encode(addressof(self.counts), counts_limit,
                                self.word_size,
                                addressof(encode_buf) + payload_header_size,
                                varint_len)

            # copy the header after updating the varint stream length
            self.payload.payload_len = varint_len
            ctypes.memmove(addressof(encode_buf), addressof(self.payload), payload_header_size)

            cdata = zlib.compress(ctypes.string_at(encode_buf, payload_header_size + varint_len))
            return cdata
        # can't compress if no payload
        raise RuntimeError('No payload to compress')

    def dump(self, label=None):
        if label:
            print 'Payload Dump ' + label
        print '   payload cookie: %x' % (self.payload.cookie)
        print '   payload_len: %d' % (self.payload.payload_len)
        print '   counts_len: %d' % (self.counts_len)
        dump_payload(self.get_counts(), self.counts_len)


class HdrHistogramEncoder(object):
    '''An encoder class for histograms, only supports V1 encoding.
    '''
    def __init__(self, histogram, b64_wrap=True, hdr_payload=None):
        '''Histogram encoder
        Args:
            histogram the histogram to encode/decode into
            b64_wrap determines if the base64 wrapper is enabled or not
            hdr_payload if None will create a new HdrPayload instance for this
                encoder, else will reuse the passed Hdrayload instance (useful
                after decoding one and to associate it to a new histogram)
            word_size counters size in bytes (2, 4 or 8)
        Exceptions:
            ValueError if the word_size value is unsupported
        '''
        self.histogram = histogram
        if not hdr_payload:
            self.payload = HdrPayload(histogram.word_size, histogram.counts_len)
            payload = self.payload.payload
            # those values never change across encodings
            payload.normalizing_index_offset = 0
            payload.conversion_ratio_bits = 1
            payload.significant_figures = histogram.significant_figures
            payload.lowest_trackable_value = histogram.lowest_trackable_value
            payload.highest_trackable_value = histogram.highest_trackable_value
        else:
            self.payload = hdr_payload

        self.b64_wrap = b64_wrap
        self.header = ExternalHeader()
        self.header.cookie = get_compression_cookie()

    def get_counts(self):
        '''Retrieve the counts array that can be used to store live counters
        and that can be encoded with minimal copies using encode()
        '''
        return self.payload.get_counts()

    def encode(self):
        '''Compress the associated encodable payload,
        prepend the header then encode with base64 if requested

        Returns:
            the b64 encoded wire encoding of the histogram (as a string)
            or the compressed payload (as a string, if b64 wrappinb is disabled)
        '''
        # only compress the first non zero buckets
        # if histogram is empty we do not encode any counter
        if self.histogram.total_count:
            relevant_length = \
                self.histogram.get_counts_array_index(self.histogram.max_value) + 1
        else:
            relevant_length = 0
        cpayload = self.payload.compress(relevant_length)
        if self.b64_wrap:
            self.header.length = len(cpayload)
            header_str = ctypes.string_at(addressof(self.header), ext_header_size)
            return base64.b64encode(''.join([header_str, cpayload]))
        return cpayload

    @staticmethod
    def decode(encoded_histogram, b64_wrap=True):
        '''Decode a wire histogram encoding into a read-only Hdr Payload instance
        Args:
            encoded_histogram a string containing the wire encoding of a histogram
                              such as one returned from encode()
        Returns:
            an hdr_payload instance with all the decoded/uncompressed fields

        Exception:
            TypeError in case of base64 decode error
            HdrCookieException:
                the main header has an invalid cookie
                the compressed payload header has an invalid cookie
            HdrLengthException:
                the decompressed size is too small for the HdrPayload structure
                or is not aligned or is too large for the passed payload class
            HdrHistogramSettingsException:
                mismatch in the significant figures, lowest and highest
                         trackable value
            zlib.error:
                in case of zlib decompression error
        '''
        if b64_wrap:
            b64decode = base64.b64decode(encoded_histogram)
            # this string has 2 parts in it: the header (raw) and the payload (compressed)
            b64dec_len = len(b64decode)

            if b64dec_len < ext_header_size:
                raise HdrLengthException('Base64 decoded message too short')

            header = ExternalHeader.from_buffer_copy(b64decode)
            if get_cookie_base(header.cookie) != V2_COMPRESSION_COOKIE_BASE:
                raise HdrCookieException()
            if header.length != b64dec_len - ext_header_size:
                raise HdrLengthException('Decoded length=%d buffer length=%d' %
                                         (header.length, b64dec_len - ext_header_size))
            # this will result in a copy of the compressed payload part
            # could not find a way to do otherwise since zlib.decompress()
            # expects a string (and does not like a buffer or a memoryview object)
            cpayload = b64decode[ext_header_size:]
        else:
            cpayload = encoded_histogram
        hdr_payload = HdrPayload(8, compressed_payload=cpayload)
        return hdr_payload

    def add(self, other_encoder):
        add_array(addressof(self.get_counts()),
                  addressof(other_encoder.get_counts()),
                  self.histogram.counts_len,
                  self.histogram.word_size)

def _dump_series(start, stop, count):
    if stop <= start + 1:
        # single index range
        print '[%06d] %d' % (start, count)
    else:
        print '[%06d] %d (%d identical)' % (start, count, stop - start)

def dump_payload(counts, max_index):
    print 'counts array size = %d entries' % (max_index)
    if not max_index:
        return
    series_start_index = 0
    total_count = 0
    current_series_count = counts[0]
    index = 0
    for index in xrange(1, max_index):
        total_count += counts[index]
        if counts[index] != current_series_count:
            # dump the current series
            _dump_series(series_start_index, index, current_series_count)
            # start a new series
            current_series_count = counts[index]
            series_start_index = index
    # there is always a last series to dump
    _dump_series(series_start_index, index, counts[index])
    print '[%06d] --END-- total count=%d' % (index + 1, total_count)

def hex_dump(label, str):
    print label
    print ':'.join(x.encode('hex') for x in str)



'''
A histogram log writer.
A Histogram logs are used to capture full fidelity, per-time-interval
histograms of a recorded value.

For example, a histogram log can be used to capture high fidelity
reaction-time logs for some measured system or subsystem component.
Such a log would capture a full reaction time histogram for each
logged interval, and could be used to later reconstruct a full
HdrHistogram of the measured reaction time behavior for any arbitrary
time range within the log, by adding [only] the relevant interval
histograms.


Ported from
https://github.com/HdrHistogram/HdrHistogram (Java)

Written by Alec Hothan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
import datetime
import re
import sys
from hdrh.histogram import HdrHistogram

class HistogramLogWriter(object):

    HISTOGRAM_LOG_FORMAT_VERSION = "1.2"

    def __init__(self, output_file):
        '''Constructs a new HistogramLogWriter that will write into the specified file.
        Params:
            output_file the File to write to
        '''
        self.log = output_file
        self.base_time = 0

    def output_interval_histogram(self,
                                  histogram,
                                  start_time_stamp_sec=0,
                                  end_time_stamp_sec=0,
                                  max_value_unit_ratio=1000000.0):
        '''Output an interval histogram, with the given timestamp and a
        configurable maxValueUnitRatio.
        (note that the specified timestamp will be used, and the timestamp in
        the actual histogram will be ignored).
        The max value reported with the interval line will be scaled by the
        given max_value_unit_ratio.
        The histogram start and end timestamps are assumed to be in msec units.
        Logging will be in seconds, realtive by a base time
        The default base time is 0.

        By covention, histogram start/end time are generally stamped with
        absolute times in msec since the epoch. For logging with absolute time
        stamps, the base time would remain zero. For
        logging with relative time stamps (time since a start point),
        Params:
            histogram The interval histogram to log.
            start_time_stamp_sec The start timestamp to log with the
                interval histogram, in seconds.
                default: using the start/end timestamp indicated in the histogram
            end_time_stamp_sec The end timestamp to log with the interval
                histogram, in seconds.
                default: using the start/end timestamp indicated in the histogram
            max_value_unit_ratio The ratio by which to divide the histogram's max
                value when reporting on it.
                default: 1,000,000 (which is the msec : nsec ratio
        '''
        if not start_time_stamp_sec:
            start_time_stamp_sec = \
                (histogram.get_start_time_stamp() - self.base_time) / 1000.0
        if not end_time_stamp_sec:
            end_time_stamp_sec = (histogram.get_end_time_stamp() - self.base_time) / 1000.0
        cpayload = histogram.encode()
        self.log.write("%f,%f,%f,%s\n" %
                       (start_time_stamp_sec,
                        end_time_stamp_sec - start_time_stamp_sec,
                        histogram.get_max_value() / max_value_unit_ratio,
                        cpayload))

    def output_start_time(self, start_time_msec):
        '''Log a start time in the log.
        Params:
            start_time_msec time (in milliseconds) since the absolute start time (the epoch)
        '''
        self.log.write("#[StartTime: %f (seconds since epoch), %s]\n" %
                       (float(start_time_msec) / 1000.0,
                        datetime.fromtimestamp(start_time_msec).iso_format(' ')))

    def output_base_time(self, base_time_msec):
        '''Log a base time in the log.
        Params:
            base_time_msec time (in milliseconds) since the absolute start time (the epoch)
        '''
        self.log.write("#[BaseTime: %f (seconds since epoch)]\n" %
                       (float(base_time_msec) / 1000.0))

    def output_comment(self, comment):
        '''Log a comment to the log.
        Comments will be preceded with with the '#' character.
        Params:
            comment the comment string.
        '''
        self.log.write("#%s\n" % (comment))

    def output_legend(self):
        '''Output a legend line to the log.
        '''
        self.log.write("\"StartTimestamp\",\"Interval_Length\","
                       "\"Interval_Max\",\"Interval_Compressed_Histogram\"\n")

    def output_log_format_version(self):
        '''Output a log format version to the log.
        '''
        self.output_comment("[Histogram log format version " +
                            HistogramLogWriter.HISTOGRAM_LOG_FORMAT_VERSION + "]")

    def close(self):
        self.log.close()

# "#[StartTime: %f (seconds since epoch), %s]\n"
re_start_time = re.compile('#\[StartTime: *([\d\.]*) ')

# "#[BaseTime: %f (seconds since epoch)]\n"
re_base_time = re.compile('#\[BaseTime: *([\d\.]*) ')

# "%f,%f,%f,%s\n"
re_histogram_interval = re.compile('([\d\.]*),([\d\.]*),([\d\.]*),(.*)')

class HistogramLogReader(object):

    def __init__(self, input_file_name, reference_histogram):
        '''Constructs a new HistogramLogReader that produces intervals read
        from the specified file name.
        Params:
            input_file_name The name of the file to read from
            reference_histogram a histogram instance used as a reference to create
                                new instances for all subsequent decoded interval
                                histograms
        '''
        self.start_time_sec = 0.0
        self.observed_start_time = False
        self.base_time_sec = 0.0
        self.observed_base_time = False
        self.input_file = open(input_file_name, "r")
        self.reference_histogram = reference_histogram

    def get_start_time_sec(self):
        '''get the latest start time found in the file so far (or 0.0),
        per the log file format explained above. Assuming the "#[StartTime:" comment
        line precedes the actual intervals recorded in the file, getStartTimeSec() can
        be safely used after each interval is read to determine's the offset of that
        interval's timestamp from the epoch.
        Return:
            latest Start Time found in the file (or 0.0 if non found)
        '''
        return self.start_time_sec

    def _decode_next_interval_histogram(self,
                                        dest_histogram,
                                        range_start_time_sec=0.0,
                                        range_end_time_sec=sys.maxint,
                                        absolute=False):
        '''Read the next interval histogram from the log, if interval falls
        within an absolute or relative time range.

        Timestamps are assumed to appear in order in the log file, and as such
        this method will return a null upon encountering a timestamp larger than
        range_end_time_sec.

        Relative time range:
            the range is assumed to be in seconds relative to
            the actual timestamp value found in each interval line in the log
        Absolute time range:
            Absolute timestamps are calculated by adding the timestamp found
            with the recorded interval to the [latest, optional] start time
            found in the log. The start time is indicated in the log with
            a "#[StartTime: " followed by the start time in seconds.

        Params:
            dest_histogram if None, created a new histogram, else adds
                           the new interval histogram to it
            range_start_time_sec The absolute or relative start of the expected
                                 time range, in seconds.
            range_start_time_sec The absolute or relative end of the expected
                                  time range, in seconds.
            absolute Defines if the passed range is absolute or relative

        Return:
            Returns an histogram object if an interval line was found with an
            associated start timestamp value that falls between start_time_sec and
            end_time_sec,
            or null if no such interval line is found.
            Upon encountering any unexpected format errors in reading the next
            interval from the file, this method will return None.

            The histogram returned will have it's timestamp set to the absolute
            timestamp calculated from adding the interval's indicated timestamp
            value to the latest [optional] start time found in the log.

        Exceptions:
            ValueError if there is a syntax error in one of the float fields
        '''
        while 1:
            line = self.input_file.readline()
            if not line:
                return None
            if line[0] == '#':
                match_res = re_start_time.match(line)
                if match_res:
                    self.start_time_sec = float(match_res.group(1))
                    self.observed_start_time = True
                    continue
                match_res = re_base_time.match(line)
                if match_res:
                    self.base_time = float(match_res.group(1))
                    self.observed_base_time = True
                    continue

            match_res = re_histogram_interval.match(line)
            if not match_res:
                # probably a legend line that starts with "\"StartTimestamp"
                continue
            # Decode: startTimestamp, intervalLength, maxTime, histogramPayload
            # Timestamp is expected to be in seconds
            log_time_stamp_in_sec = float(match_res.group(1))
            interval_length_sec = float(match_res.group(2))
            cpayload = match_res.group(4)

            if not self.observed_start_time:
                # No explicit start time noted. Use 1st observed time:
                self.start_time_sec = log_time_stamp_in_sec
                self.observed_start_time = True

            if not self.observed_base_time:
                # No explicit base time noted.
                # Deduce from 1st observed time (compared to start time):
                if log_time_stamp_in_sec < self.start_time_sec - (365 * 24 * 3600.0):
                    # Criteria Note: if log timestamp is more than a year in
                    # the past (compared to StartTime),
                    # we assume that timestamps in the log are not absolute
                    self.base_time_sec = self.start_time_sec
                else:
                    # Timestamps are absolute
                    self.base_time_sec = 0.0
                self.observed_base_time = True

            absolute_start_time_stamp_sec = \
                log_time_stamp_in_sec + self.base_time_sec
            offset_start_time_stamp_sec = \
                absolute_start_time_stamp_sec - self.start_time_sec

            # Timestamp length is expect to be in seconds
            absolute_end_time_stamp_sec = \
                absolute_start_time_stamp_sec + interval_length_sec

            if absolute:
                start_time_stamp_to_check_range_on = absolute_start_time_stamp_sec
            else:
                start_time_stamp_to_check_range_on = offset_start_time_stamp_sec

            if start_time_stamp_to_check_range_on < range_start_time_sec:
                continue

            if start_time_stamp_to_check_range_on > range_end_time_sec:
                return None
            if dest_histogram:
                # add the interval histogram to the destination histogram
                histogram = dest_histogram
                histogram.decode_and_add(cpayload)
            else:
                histogram = HdrHistogram.decode(cpayload)
                histogram.set_start_time_stamp(absolute_start_time_stamp_sec * 1000.0)
                histogram.set_end_time_stamp(absolute_end_time_stamp_sec * 1000.0)
            return histogram

    def get_next_interval_histogram(self,
                                    range_start_time_sec=0.0,
                                    range_end_time_sec=sys.maxint,
                                    absolute=False):
        '''Read the next interval histogram from the log, if interval falls
        within an absolute or relative time range.

        Timestamps are assumed to appear in order in the log file, and as such
        this method will return a null upon encountering a timestamp larger than
        range_end_time_sec.

        Relative time range:
            the range is assumed to be in seconds relative to
            the actual timestamp value found in each interval line in the log
        Absolute time range:
            Absolute timestamps are calculated by adding the timestamp found
            with the recorded interval to the [latest, optional] start time
            found in the log. The start time is indicated in the log with
            a "#[StartTime: " followed by the start time in seconds.

        Params:

            range_start_time_sec The absolute or relative start of the expected
                                 time range, in seconds.
            range_start_time_sec The absolute or relative end of the expected
                                  time range, in seconds.
            absolute Defines if the passed range is absolute or relative

        Return:
            Returns an histogram object if an interval line was found with an
            associated start timestamp value that falls between start_time_sec and
            end_time_sec,
            or null if no such interval line is found.
            Upon encountering any unexpected format errors in reading the next
            interval from the file, this method will return None.

            The histogram returned will have it's timestamp set to the absolute
            timestamp calculated from adding the interval's indicated timestamp
            value to the latest [optional] start time found in the log.

        Exceptions:
            ValueError if there is a syntax error in one of the float fields
        '''
        return self._decode_next_interval_histogram(None,
                                                    range_start_time_sec,
                                                    range_end_time_sec,
                                                    absolute)

    def add_next_interval_histogram(self,
                                    dest_histogram=None,
                                    range_start_time_sec=0.0,
                                    range_end_time_sec=sys.maxint,
                                    absolute=False):
        '''Read the next interval histogram from the log, if interval falls
        within an absolute or relative time range, and add it to the destination
        histogram (or to the reference histogram if dest_histogram is None)

        Timestamps are assumed to appear in order in the log file, and as such
        this method will return a null upon encountering a timestamp larger than
        range_end_time_sec.

        Relative time range:
            the range is assumed to be in seconds relative to
            the actual timestamp value found in each interval line in the log
        Absolute time range:
            Absolute timestamps are calculated by adding the timestamp found
            with the recorded interval to the [latest, optional] start time
            found in the log. The start time is indicated in the log with
            a "#[StartTime: " followed by the start time in seconds.

        Params:
            dest_histogram where to add the next interval histogram, if None
                           the interal histogram will be added to the reference
                           histogram passed in the constructor
            range_start_time_sec The absolute or relative start of the expected
                                 time range, in seconds.
            range_start_time_sec The absolute or relative end of the expected
                                  time range, in seconds.
            absolute Defines if the passed range is absolute or relative

        Return:
            Returns the destination histogram if an interval line was found with an
            associated start timestamp value that falls between start_time_sec and
            end_time_sec,
            or None if no such interval line is found.
            Upon encountering any unexpected format errors in reading the next
            interval from the file, this method will return None.

            The histogram returned will have it's timestamp set to the absolute
            timestamp calculated from adding the interval's indicated timestamp
            value to the latest [optional] start time found in the log.

        Exceptions:
            ValueError if there is a syntax error in one of the float fields
        '''
        if not dest_histogram:
            dest_histogram = self.reference_histogram
        return self._decode_next_interval_histogram(dest_histogram,
                                                    range_start_time_sec,
                                                    range_end_time_sec,
                                                    absolute)

    def close(self):
        self.input_file.close()



import pytest
def pytest_addoption(parser):
    parser.addoption("--runperf", action="store_true",
        help="run perf tests")

def pytest_runtest_setup(item):
    if 'perf' in item.keywords and not item.config.getoption("--runperf"):
        pytest.skip("need --runperf option to run")


'''
A pure python version of the hdr_histogram code

Ported from
https://github.com/HdrHistogram/HdrHistogram (Java)
https://github.com/HdrHistogram/HdrHistogram_c (C)

Histogram Iterators: all values, recorded, linear and logarithmic

Written by Alec Hothan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
from abc import abstractmethod
import math

class HdrConcurrentModificationException(Exception):
    pass

class HdrIterationValue(object):
    '''Class of the values returned by each iterator
    '''
    def __init__(self, hdr_iterator):
        self.hdr_iterator = hdr_iterator
        self.value_iterated_to = 0
        self.value_iterated_from = 0,
        self.count_at_value_iterated_to = 0
        self.count_added_in_this_iter_step = 0
        self.total_count_to_this_value = 0
        self.total_value_to_this_value = 0
        self.percentile = 0.0
        self.percentile_level_iterated_to = 0.0
        self.int_to_double_conversion_ratio = 0.0

    def set(self, value_iterated_to):
        hdr_it = self.hdr_iterator
        self.value_iterated_to = value_iterated_to
        self.value_iterated_from = hdr_it.prev_value_iterated_to
        self.count_at_value_iterated_to = hdr_it.count_at_this_value
        self.count_added_in_this_iter_step = \
            hdr_it.total_count_to_current_index - hdr_it.total_count_to_prev_index
        self.total_count_to_this_value = hdr_it.total_count_to_current_index
        self.total_value_to_this_value = hdr_it.value_to_index
        self.percentile = (100.0 * hdr_it.total_count_to_current_index) / hdr_it.total_count
        self.percentile_level_iterated_to = hdr_it.get_percentile_iterated_to()
        self.int_to_double_conversion_ratio = hdr_it.int_to_double_conversion_ratio

class AbstractHdrIterator(object):
    '''Provide a means of iterating through all histogram values using the finest
    granularity steps supported by the underlying representation.
    The iteration steps through all possible unit value levels, regardless of
    whether or not there were recorded values for that value level,
    and terminates when all recorded histogram values are exhausted.
    '''
    def __init__(self, histogram):
        self.histogram = histogram
        self.current_index = 0
        self.count_at_this_value = 0
        self.total_count_to_current_index = 0
        self.total_count_to_prev_index = 0
        self.prev_value_iterated_to = 0
        self.value_at_index = 0
        self.value_to_index = 0
        self.value_at_next_index = 1 << histogram.unit_magnitude
        self.current_iteration_value = HdrIterationValue(self)
        # take a snapshot of the total count
        self.total_count = histogram.total_count
        self.int_to_double_conversion_ratio = histogram.int_to_double_conversion_ratio
        self.fresh_sub_bucket = True

    def __iter__(self):
        self.reset_iterator(self.histogram)
        return self

    def reset_iterator(self, histogram):
        if not histogram:
            histogram = self.histogram
        self.histogram = histogram
        self.current_index = 0
        self.count_at_this_value = 0
        self.total_count_to_current_index = 0
        self.total_count_to_prev_index = 0
        self.prev_value_iterated_to = 0
        self.value_at_index = 0
        self.value_to_index = 0
        self.value_at_next_index = 1 << histogram.unit_magnitude
        self.current_iteration_value = HdrIterationValue(self)
        # take a snapshot of the total count
        self.total_count = histogram.total_count
        self.int_to_double_conversion_ratio = histogram.int_to_double_conversion_ratio
        self.fresh_sub_bucket = True

    def has_next(self):
        return self.total_count_to_current_index < self.total_count

    def next(self):
        if self.total_count != self.histogram.total_count:
            raise HdrConcurrentModificationException()
        while self.has_next():
            self.count_at_this_value = self.histogram.get_count_at_index(self.current_index)
            if self.fresh_sub_bucket:
                self.total_count_to_current_index += self.count_at_this_value
                self.value_to_index += self.count_at_this_value * self.get_value_iterated_to()
                self.fresh_sub_bucket = False
            if self.reached_iteration_level():
                value_iterated_to = self.get_value_iterated_to()
                self.current_iteration_value.set(value_iterated_to)

                self.prev_value_iterated_to = value_iterated_to
                self.total_count_to_prev_index = self.total_count_to_current_index

                self.increment_iteration_level()

                if self.total_count != self.histogram.total_count:
                    raise HdrConcurrentModificationException()

                return self.current_iteration_value

            # get to the next sub bucket
            self.increment_sub_bucket()

        if self.total_count_to_current_index > self.total_count_to_prev_index:
            # We are at the end of the iteration but we still need to report
            # the last iteration value
            value_iterated_to = self.get_value_iterated_to()
            self.current_iteration_value.set(value_iterated_to)
            # we do this one time only
            self.total_count_to_prev_index = self.total_count_to_current_index
            return self.current_iteration_value

        raise StopIteration()

    @abstractmethod
    def reached_iteration_level(self):
        pass

    @abstractmethod
    def increment_iteration_level(self):
        pass

    def increment_sub_bucket(self):
        self.fresh_sub_bucket = True
        self.current_index += 1
        self.value_at_index = self.histogram.get_value_from_index(self.current_index)
        self.value_at_next_index = \
            self.histogram.get_value_from_index(self.current_index + 1)

    def get_value_iterated_to(self):
        return self.histogram.get_highest_equivalent_value(self.value_at_index)

    def get_percentile_iterated_to(self):
        return (100.0 * self.total_count_to_current_index) / self.total_count

    def get_percentile_iterated_from(self):
        return (100.0 * self.total_count_to_prev_index) / self.total_count

class AllValuesIterator(AbstractHdrIterator):
    def __init__(self, histogram):
        AbstractHdrIterator.__init__(self, histogram)
        self.visited_index = -1

    def reset(self, histogram=None):
        AbstractHdrIterator.reset_iterator(self, histogram)
        self.visited_index = -1

    def reached_iteration_level(self):
        return self.visited_index != self.current_index

    def increment_iteration_level(self):
        self.visited_index = self.current_index

    def has_next(self):
        if self.total_count != self.histogram.total_count:
            raise HdrConcurrentModificationException()
        return self.current_index < self.histogram.counts_len - 1

class RecordedIterator(AllValuesIterator):
    '''Provide a means of iterating through all recorded histogram values
    using the finest granularity steps supported by the underlying representation.
    The iteration steps through all non-zero recorded value counts,
    and terminates when all recorded histogram values are exhausted.
    '''
    def reached_iteration_level(self):
        current_count = self.histogram.get_count_at_index(self.current_index)
        return current_count and self.visited_index != self.current_index

class AbstractLiLoIteratortype(AbstractHdrIterator):
    '''Linear/Log iterator common parent class
    '''
    def __init__(self, histogram, next_value_report_lev):
        AbstractHdrIterator.__init__(self, histogram)
        self.next_value_report_lev = next_value_report_lev
        self.next_value_report_lev_lowest_eq = \
            histogram.get_lowest_equivalent_value(next_value_report_lev)

    def reset(self, histogram, next_value_report_lev):
        AbstractHdrIterator.reset_iterator(self, histogram)
        if next_value_report_lev:
            self.next_value_report_lev = next_value_report_lev
        self.next_value_report_lev_lowest_eq = \
            self.histogram.get_lowest_equivalent_value(next_value_report_lev)

    def has_next(self):
        if AbstractHdrIterator.has_next(self):
            return True
        # If next iterate does not move to the next sub bucket index (which is empty if
        # if we reached this point), then we are not done iterating... Otherwise we're done.
        return self.next_value_report_lev_lowest_eq < self.value_at_next_index

    def reached_iteration_level(self):
        return self.value_at_index >= self.next_value_report_lev_lowest_eq

    def get_value_iterated_to(self):
        return self.next_value_report_lev

class LinearIterator(AbstractLiLoIteratortype):
    '''Provide a means of iterating through histogram values using linear steps.

    The iteration is performed in steps of value_units_per_bucket in size,
    terminating when all recorded histogram values are exhausted.
    '''
    def __init__(self, histogram, value_units_per_bucket):
        AbstractLiLoIteratortype.__init__(self, histogram, value_units_per_bucket)
        self.value_units_per_bucket = value_units_per_bucket

    def reset(self, histogram=None, value_units_per_bucket=0):
        AbstractLiLoIteratortype.reset(self, histogram, value_units_per_bucket)
        if value_units_per_bucket:
            self.value_units_per_bucket = value_units_per_bucket

    def increment_iteration_level(self):
        self.next_value_report_lev += self.value_units_per_bucket
        self.next_value_report_lev_lowest_eq = \
            self.histogram.get_lowest_equivalent_value(self.next_value_report_lev)

class LogIterator(AbstractLiLoIteratortype):
    '''Provide a means of iterating through histogram values at logarithmically
    increasing levels.

    The iteration is performed in steps that start at value_units_first_bucket
    and increase exponentially according to log_base, terminating when all
    recorded histogram values are exhausted.
    '''
    def __init__(self, histogram, value_units_first_bucket, log_base):
        AbstractLiLoIteratortype.__init__(self, histogram, value_units_first_bucket)
        self.log_base = log_base

    def reset(self, histogram=None, value_units_first_bucket=0, log_base=0):
        AbstractLiLoIteratortype.reset(self, histogram, value_units_first_bucket)
        if log_base:
            self.log_base = log_base

    def increment_iteration_level(self):
        self.next_value_report_lev *= self.log_base
        self.next_value_report_lev_lowest_eq = \
            self.histogram.get_lowest_equivalent_value(self.next_value_report_lev)

class PercentileIterator(AbstractHdrIterator):
    '''Provide a means of iterating through histogram values according to
    percentile levels

    The iteration is performed in steps that start at 0% and reduce their
    distance to 100% according to the ticks_per_half_distance parameter,
    ultimately reaching 100% when all recorded histogram values are exhausted.
    '''
    def __init__(self, histogram, percentile_ticks_per_half_distance):
        AbstractHdrIterator.__init__(self, histogram)
        self.percentile_ticks_per_half_distance = percentile_ticks_per_half_distance
        self.percentile_to_iterate_to = 0
        self.percentile_to_iterate_from = 0
        self.reached_last_recorded_value = False

    def reset(self, histogram=None, percentile_ticks_per_half_distance=0):
        AbstractHdrIterator.reset_iterator(self, histogram)
        if percentile_ticks_per_half_distance:
            self.percentile_ticks_per_half_distance = percentile_ticks_per_half_distance
        self.percentile_to_iterate_to = 0
        self.percentile_to_iterate_from = 0
        self.reached_last_recorded_value = False

    def has_next(self):
        if AbstractHdrIterator.has_next(self):
            return True

        # We want one additional last step to 100%
        if not self.reached_last_recorded_value and self.total_count:
            self.percentile_to_iterate_to = 100.0
            self.reached_last_recorded_value = True
            return True
        return False

    def increment_iteration_level(self):
        self.percentile_to_iterate_from = self.percentile_to_iterate_to
        percentile_gap = 100.0 - (self.percentile_to_iterate_to)
        if percentile_gap:
            half_distance = math.pow(2,
                                     (math.log(100 / percentile_gap) / math.log(2)) + 1)
            percentile_reporting_ticks = self.percentile_ticks_per_half_distance * half_distance
            self.percentile_to_iterate_to += 100.0 / percentile_reporting_ticks

    def reached_iteration_level(self):
        if self.count_at_this_value == 0:
            return False
        current_percentile = (100.0 * self.total_count_to_current_index) / self.total_count
        return current_percentile >= self.percentile_to_iterate_to

    def get_percentile_iterated_to(self):
        return self.percentile_to_iterate_to

    def get_percentile_iterated_from(self):
        return self.percentile_to_iterate_from



import sys

from setuptools import setup
from setuptools import Extension
from setuptools.command.test import test


class Tox(test):
    def initialize_options(self):
        test.initialize_options(self)
        self.tox_args = None

    def finalize_options(self):
        test.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import tox
        sys.exit(tox.cmdline())

if __name__ == '__main__':
    setup(setup_requires=['pbr'], pbr=True,
          keywords='hdrhistogram hdr histogram high dynamic range',
          tests_require=['tox'],
          cmdclass={'test': Tox},
          ext_modules = [Extension('pyhdrh',
                                   sources = ['src/python-codec.c'])]
    )



'''
Test code for the python version of HdrHistogram.

Ported from
https://github.com/HdrHistogram/HdrHistogram (Java)
https://github.com/HdrHistogram/HdrHistogram_c (C)

Written by Alec Hothan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
import cProfile
import datetime
import pytest
import zlib

from ctypes import addressof
from ctypes import c_uint8
from ctypes import c_uint16
from ctypes import c_uint32
from ctypes import c_uint64
from ctypes import sizeof
from ctypes import string_at

from hdrh.codec import HdrPayload
from hdrh.codec import HdrCookieException
from hdrh.histogram import HdrHistogram
from hdrh.log import HistogramLogWriter
from hdrh.log import HistogramLogReader
from pyhdrh import add_array
from pyhdrh import encode
from pyhdrh import decode

# histogram __init__ values
LOWEST = 1
HIGHEST = 3600 * 1000 * 1000
SIGNIFICANT = 3
TEST_VALUE_LEVEL = 4
INTERVAL = 10000

@pytest.mark.basic
def test_basic():
    histogram = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    assert(histogram.bucket_count == 22)
    assert(histogram.sub_bucket_count == 2048)
    assert(histogram.counts_len == 23552)
    assert(histogram.unit_magnitude == 0)
    assert(histogram.sub_bucket_half_count_magnitude == 10)

@pytest.mark.basic
def test_empty_histogram():
    histogram = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    assert(histogram.get_min_value() == 0)
    assert(histogram.get_max_value() == 0)
    assert(histogram.get_mean_value() == 0)
    assert(histogram.get_stddev() == 0)

@pytest.mark.basic
def test_large_numbers():
    histogram = HdrHistogram(20000000, 100000000, 5)
    histogram.record_value(100000000)
    histogram.record_value(20000000)
    histogram.record_value(30000000)
    assert(histogram.values_are_equivalent(20000000, histogram.get_value_at_percentile(50.0)))
    assert(histogram.values_are_equivalent(30000000, histogram.get_value_at_percentile(83.33)))
    assert(histogram.values_are_equivalent(100000000, histogram.get_value_at_percentile(83.34)))
    assert(histogram.values_are_equivalent(100000000, histogram.get_value_at_percentile(99.0)))

@pytest.mark.basic
def test_record_value():
    histogram = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    histogram.record_value(TEST_VALUE_LEVEL)
    assert(histogram.get_count_at_value(TEST_VALUE_LEVEL) == 1)
    assert(histogram.get_total_count() == 1)

@pytest.mark.basic
def test_highest_equivalent_value():
    histogram = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    assert 8183 * 1024 + 1023 == histogram.get_highest_equivalent_value(8180 * 1024)
    assert 8191 * 1024 + 1023 == histogram.get_highest_equivalent_value(8191 * 1024)
    assert 8199 * 1024 + 1023 == histogram.get_highest_equivalent_value(8193 * 1024)
    assert 9999 * 1024 + 1023 == histogram.get_highest_equivalent_value(9995 * 1024)
    assert 10007 * 1024 + 1023 == histogram.get_highest_equivalent_value(10007 * 1024)
    assert 10015 * 1024 + 1023 == histogram.get_highest_equivalent_value(10008 * 1024)

@pytest.mark.basic
def test_scaled_highest_equiv_value():
    histogram = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    assert 8183 == histogram.get_highest_equivalent_value(8180)
    assert 8191 == histogram.get_highest_equivalent_value(8191)
    assert 8199 == histogram.get_highest_equivalent_value(8193)
    assert 9999 == histogram.get_highest_equivalent_value(9995)
    assert 10007 == histogram.get_highest_equivalent_value(10007)
    assert 10015 == histogram.get_highest_equivalent_value(10008)

def load_histogram():
    histogram = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    # record this value with a count of 10,000
    histogram.record_value(1000L, 10000)
    histogram.record_value(100000000L)
    return histogram

def load_corrected_histogram():
    histogram = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    # record this value with a count of 10,000
    histogram.record_corrected_value(1000L, INTERVAL, 10000)
    histogram.record_corrected_value(100000000L, INTERVAL)
    return histogram

def check_percentile(hist, percentile, value, variation):
    value_at = hist.get_value_at_percentile(percentile)
    assert(abs(value_at - value) < value * variation)

def check_hist_percentiles(hist, total_count, perc_value_list):
    for pair in perc_value_list:
        check_percentile(hist, pair[0], pair[1], 0.001)
    assert(hist.get_total_count() == total_count)
    assert(hist.values_are_equivalent(hist.get_min_value(), 1000.0))
    assert(hist.values_are_equivalent(hist.get_max_value(), 100000000.0))

def check_percentiles(histogram, corrected_histogram):
    check_hist_percentiles(histogram,
                           10001,
                           ((30.0, 1000.0),
                            (99.0, 1000.0),
                            (99.99, 1000.0),
                            (99.999, 100000000.0),
                            (100.0, 100000000.0)))
    check_hist_percentiles(corrected_histogram,
                           20000,
                           ((30.0, 1000.0),
                            (50.0, 1000.0),
                            (75.0, 50000000.0),
                            (90.0, 80000000.0),
                            (99.0, 98000000.0),
                            (99.999, 100000000.0),
                            (100.0, 100000000.0)))

@pytest.mark.basic
def test_percentiles():
    check_percentiles(load_histogram(), load_corrected_histogram())

@pytest.mark.iterators
def test_recorded_iterator():
    hist = load_histogram()
    index = 0
    for item in hist.get_recorded_iterator():
        count_added_in_this_bucket = item.count_added_in_this_iter_step
        if index == 0:
            assert(count_added_in_this_bucket == 10000)
        else:
            assert(count_added_in_this_bucket == 1)
        index += 1
    assert(index == 2)

    hist = load_corrected_histogram()
    index = 0
    total_added_count = 0
    for item in hist.get_recorded_iterator():
        count_added_in_this_bucket = item.count_added_in_this_iter_step
        if index == 0:
            assert(count_added_in_this_bucket == 10000)

        assert(item.count_at_value_iterated_to != 0)
        total_added_count += count_added_in_this_bucket
        index += 1
    assert(total_added_count == 20000)
    assert(total_added_count == hist.get_total_count())

def check_iterator_values(itr, last_index):
    index = 0
    for item in itr:
        count_added_in_this_bucket = item.count_added_in_this_iter_step
        if index == 0:
            assert(count_added_in_this_bucket == 10000)
        elif index == last_index:
            assert(count_added_in_this_bucket == 1)
        else:
            assert(count_added_in_this_bucket == 0)
        index += 1
    assert(index - 1 == last_index)

def check_corrected_iterator_values(itr, last_index):
    index = 0
    total_added_count = 0
    for item in itr:
        count_added_in_this_bucket = item.count_added_in_this_iter_step
        if index == 0:
            # first bucket is range [0, 10000]
            # value 1000  count = 10000
            # value 10000 count = 1 (corrected from the 100M value with 10K interval)
            assert(count_added_in_this_bucket == 10001)
        index += 1
        total_added_count += count_added_in_this_bucket

    assert(index - 1 == last_index)
    assert(total_added_count == 20000)

@pytest.mark.iterators
def test_linear_iterator():
    hist = load_histogram()
    itr = hist.get_linear_iterator(100000)
    check_iterator_values(itr, 999)
    hist = load_corrected_histogram()
    itr = hist.get_linear_iterator(10000)
    check_corrected_iterator_values(itr, 9999)

@pytest.mark.iterators
def test_log_iterator():
    hist = load_histogram()
    itr = hist.get_log_iterator(10000, 2.0)
    check_iterator_values(itr, 14)
    hist = load_corrected_histogram()
    itr = hist.get_log_iterator(10000, 2.0)
    check_corrected_iterator_values(itr, 14)

@pytest.mark.iterators
def test_percentile_iterator():
    hist = load_histogram()
    # test with 5 ticks per half distance
    for item in hist.get_percentile_iterator(5):
        expected = hist.get_highest_equivalent_value(hist.get_value_at_percentile(item.percentile))
        assert(item.value_iterated_to == expected)

@pytest.mark.iterators
def test_reset_iterator():
    hist = load_corrected_histogram()
    itr = hist.get_recorded_iterator()

    # do a partial iteration
    index = 0
    total_added_count = 0
    for item in itr:
        count_added_in_this_bucket = item.count_added_in_this_iter_step
        if index == 0:
            assert(count_added_in_this_bucket == 10000)

        assert(item.count_at_value_iterated_to != 0)
        total_added_count += count_added_in_this_bucket
        index += 1
        if total_added_count >= 10000:
            break

    # reset iterator and do a full iteration
    itr.reset()
    index = 0
    total_added_count = 0
    for item in itr:
        count_added_in_this_bucket = item.count_added_in_this_iter_step
        if index == 0:
            assert(count_added_in_this_bucket == 10000)

        assert(item.count_at_value_iterated_to != 0)
        total_added_count += count_added_in_this_bucket
        index += 1
    assert(total_added_count == 20000)
    assert(total_added_count == hist.get_total_count())

    # just run the reset method
    hist.get_all_values_iterator().reset()
    hist.get_linear_iterator(100000).reset()
    hist.get_log_iterator(10000, 2.0).reset()
    hist.get_percentile_iterator(5).reset()

@pytest.mark.basic
def test_reset():
    histogram = load_histogram()
    histogram.reset()
    assert(histogram.get_total_count() == 0)
    assert(histogram.get_value_at_percentile(99.99) == 0)

@pytest.mark.basic
def test_invalid_significant_figures():
    try:
        HdrHistogram(LOWEST, HIGHEST, -1)
        assert(False)
    except ValueError:
        pass
    try:
        HdrHistogram(LOWEST, HIGHEST, 6)
        assert(False)
    except ValueError:
        pass

@pytest.mark.basic
def test_out_of_range_values():
    histogram = HdrHistogram(1, 1000, 4)
    assert(histogram.record_value(32767))
    assert(histogram.record_value(32768) is False)

# Make up a list of values for testing purpose
VALUES_LIST = (
    1000,
    1000,
    3000,
    3000
)

@pytest.mark.basic
def test_mean_stddev():
    # fill up a histogram with the values in the list
    histogram = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    for value in VALUES_LIST:
        histogram.record_value(value)
    assert(histogram.get_mean_value() == 2000.5)
    assert(histogram.get_stddev() == 1000.5)

HDR_PAYLOAD_COUNTS = 1000
HDR_PAYLOAD_PARTIAL_COUNTS = HDR_PAYLOAD_COUNTS / 2

def fill_counts(payload, last_index, start=0):
    # note that this function should only be used for
    # raw payload level operations, shoud not be used for payloads that are
    # created from a histogram, see fill_hist_counts
    counts = payload.get_counts()
    for index in xrange(start, last_index):
        counts[index] = index

def check_counts(payload, last_index, multiplier=1, start=0):
    counts = payload.get_counts()
    for index in xrange(start, last_index):
        assert(counts[index] == multiplier * index)

def check_hdr_payload(counter_size):
    # Create an HdrPayload class with given counters count
    payload = HdrPayload(counter_size, HDR_PAYLOAD_COUNTS)
    # put some known numbers in the buckets
    fill_counts(payload, HDR_PAYLOAD_COUNTS)

    # get a compressed version of that payload
    cpayload = payload.compress(HDR_PAYLOAD_COUNTS)
    # now decompress it into a new hdr payload instance
    dpayload = HdrPayload(counter_size, compressed_payload=cpayload)
    dpayload.init_counts(HDR_PAYLOAD_COUNTS)

    # now verify that the counters are identical to the original
    check_counts(dpayload, HDR_PAYLOAD_COUNTS)

@pytest.mark.codec
def test_hdr_payload():
    # Check the payload work in all 3 supported counter sizes
    for counter_size in [2, 4, 8]:
        check_hdr_payload(counter_size)

@pytest.mark.codec
def test_hdr_payload_exceptions():
    # test invalid zlib compressed buffer
    with pytest.raises(zlib.error):
        HdrPayload(2, compressed_payload="junk data")

    # unsupported word size
    with pytest.raises(ValueError):
        payload = HdrPayload(1, HDR_PAYLOAD_COUNTS)
    with pytest.raises(ValueError):
        payload = HdrPayload(1000, HDR_PAYLOAD_COUNTS)

    # invalid cookie
    payload = HdrPayload(8, HDR_PAYLOAD_COUNTS)
    payload.payload.cookie = 12345
    cpayload = payload.compress(HDR_PAYLOAD_COUNTS)
    with pytest.raises(HdrCookieException):
        HdrPayload(2, compressed_payload=cpayload)

def fill_hist_counts(histogram, last_index, start=0):
    # fill the counts of a given histogram and update the min/max/total count
    # accordingly
    for index in xrange(start, last_index):
        value_from_index = histogram.get_value_from_index(index)
        histogram.record_value(value_from_index, index)

def check_hist_counts(histogram, last_index, multiplier=1, start=0):
    for index in xrange(start, last_index):
        assert(histogram.get_count_at_index(index) == multiplier * index)

# This is the max latency used by wrk2
WRK2_MAX_LATENCY = 24 * 60 * 60 * 1000000

def check_hist_encode(word_size,
                      digits,
                      expected_compressed_length,
                      fill_start_percent,
                      fill_count_percent):
    histogram = HdrHistogram(LOWEST, WRK2_MAX_LATENCY, digits,
                             word_size=word_size)
    if fill_count_percent:
        fill_start_index = (fill_start_percent * histogram.counts_len) / 100
        fill_to_index = fill_start_index + (fill_count_percent * histogram.counts_len) / 100
        fill_hist_counts(histogram, fill_to_index, fill_start_index)
    b64 = histogram.encode()
    assert(len(b64) == expected_compressed_length)

# A list of call arguments to check_hdr_encode
ENCODE_ARG_LIST = (
    # word size digits  expected_compressed_length, fill_start%, fill_count%
    # best case when all counters are zero
    (8, 3, 48, 0, 0),        # V1=52 385 = size when compressing entire counts array
    (8, 2, 48, 0, 0),        # 126
    # typical case when all counters are aggregated in a small contiguous area
    (8, 3, 15560, 30, 20),   # V1=16452
    (8, 2, 1688, 30, 20),    # V1=2096
    # worst case when all counters are different
    (8, 3, 76892, 0, 100),   # V1=80680
    (8, 2, 9340, 0, 100),    # V1=10744
    # worst case 32-bit and 16-bit counters
    (2, 3, 76892, 0, 100),   # V1=68936
    (2, 2, 9340, 0, 100),    # V1=9144
)

@pytest.mark.codec
def test_hist_encode():
    for args in ENCODE_ARG_LIST:
        check_hist_encode(*args)

@pytest.mark.codec
def check_hist_codec_b64(word_size, b64_wrap):
    histogram = HdrHistogram(LOWEST, WRK2_MAX_LATENCY, SIGNIFICANT,
                             b64_wrap=b64_wrap,
                             word_size=word_size)
    # encode with all zero counters
    encoded = histogram.encode()
    # add back same histogram
    histogram.decode_and_add(encoded)
    # counters should remain zero
    check_hist_counts(histogram, histogram.counts_len, multiplier=0)
    # fill up the histogram
    fill_hist_counts(histogram, histogram.counts_len)
    encoded = histogram.encode()
    histogram.decode_and_add(encoded)
    check_hist_counts(histogram, histogram.counts_len, multiplier=2)

@pytest.mark.codec
def test_hist_codec():
    for word_size in [2, 4, 8]:
        check_hist_codec_b64(word_size, True)
        check_hist_codec_b64(word_size, False)

@pytest.mark.codec
def test_hist_codec_partial():
    histogram = HdrHistogram(LOWEST, WRK2_MAX_LATENCY, SIGNIFICANT)

    partial_histogram = HdrHistogram(LOWEST, WRK2_MAX_LATENCY, SIGNIFICANT)

    # put some known numbers in the first half buckets
    half_count = partial_histogram.counts_len
    fill_hist_counts(partial_histogram, half_count)
    encoded = partial_histogram.encode()
    histogram.decode_and_add(encoded)

    # now verify that the partial counters are identical to the original
    check_hist_counts(histogram, half_count, multiplier=1)
    check_hist_counts(histogram, histogram.counts_len, start=half_count + 1, multiplier=0)

# A list of encoded histograms as generated by the test code in HdrHistogram_c
# encoded from the standard Hdr test histograms (load_histogram())
# These are all histograms with 64-bit counters

ENCODE_SAMPLES_HDRHISTOGRAM_C = [
    # standard Hdr test histogram
    'HISTFAAAACl4nJNpmSzMwMBgyAABzFCaEURcm7yEwf4DROA8/4I5jNM7mJgAlWkH9g==',
    # standard Hdr test corrected histogram
    'HISTFAAAAP94nJNpmSzMwCByigECmKE0I4i4NnkJg/0HiMB5/gVzGD8aM/3lZ7rPyTSbjektC9N7Fqa'
    'HzEzbmZi2whCEvZKRaSYj02wwiYng4tFM3lDoC2dhhwh5UyZlJlUMjClCmgpMEUUmQSZ+IBZEojFFCM'
    'vQRwUxenmZ2MGQFUqz4+CTo4I2pg4dFdQylZWJkYkZCaPyMEUIydPLjMHrssFixuB12XD0HRAwMsFJg'
    'kwilZHJHHKmjzp41MFDOjhYmFiQEUEmMWqGsvKBd8Fwd/Co/waTC9jYOMAIiSJRhLbKh5x9Q87Bo/YN'
    'bfsoM4CPhw+IIJAkxnDXN+QcTIpyAPnGh6k='
]

@pytest.mark.codec
def test_hdr_interop():
    # decode and add the encoded histograms
    histogram = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    corrected_histogram = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    histogram.decode_and_add(ENCODE_SAMPLES_HDRHISTOGRAM_C[0])
    corrected_histogram.decode_and_add(ENCODE_SAMPLES_HDRHISTOGRAM_C[1])

    # check the percentiles. min, max values match
    check_percentiles(histogram, corrected_histogram)

def check_cod_perf():
    histogram = HdrHistogram(LOWEST, WRK2_MAX_LATENCY, 2)
    fill_start_index = (20 * histogram.counts_len) / 100
    fill_to_index = fill_start_index + (30 * histogram.counts_len) / 100
    fill_hist_counts(histogram, fill_to_index, fill_start_index)

    # encode 1000 times
    start = datetime.datetime.now()
    for _ in xrange(1000):
        histogram.encode()
    delta = datetime.datetime.now() - start
    print delta

def check_dec_perf():
    histogram = HdrHistogram(LOWEST, WRK2_MAX_LATENCY, 2)
    fill_start_index = (20 * histogram.counts_len) / 100
    fill_to_index = fill_start_index + (30 * histogram.counts_len) / 100
    fill_hist_counts(histogram, fill_to_index, fill_start_index)
    b64 = histogram.encode()

    # decode and add to self 1000 times
    start = datetime.datetime.now()
    for _ in xrange(1000):
        histogram.decode_and_add(b64)
    delta = datetime.datetime.now() - start
    print delta

@pytest.mark.perf
def test_cod_perf():
    cProfile.runctx('check_cod_perf()', globals(), locals())

@pytest.mark.perf
def test_dec_perf():
    cProfile.runctx('check_dec_perf()', globals(), locals())

def check_decoded_hist_counts(hist, multiplier):
    assert(hist)
    check_hist_counts(hist, hist.counts_len, multiplier)

HDR_LOG_NAME = 'hdr.log'
@pytest.mark.log
def test_log():
    # 3 histograms instances with various content
    empty_hist = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    hist = load_histogram()
    corrected_hist = load_corrected_histogram()
    with open(HDR_LOG_NAME, 'w') as hdr_log:
        log_writer = HistogramLogWriter(hdr_log)
        log_writer.output_comment("Logged with hdrhistogram.py")
        log_writer.output_log_format_version()
        log_writer.output_legend()
        # snapshot the 3 histograms
        log_writer.output_interval_histogram(empty_hist)
        log_writer.output_interval_histogram(hist)
        log_writer.output_interval_histogram(corrected_hist)
        log_writer.close()

    # decode the log file and check the decoded histograms
    log_reader = HistogramLogReader(HDR_LOG_NAME, empty_hist)
    decoded_empty_hist = log_reader.get_next_interval_histogram()
    check_decoded_hist_counts(decoded_empty_hist, 0)
    decoded_hist = log_reader.get_next_interval_histogram()
    decoded_corrected_hist = log_reader.get_next_interval_histogram()
    check_percentiles(decoded_hist, decoded_corrected_hist)
    assert(log_reader.get_next_interval_histogram() is None)

JHICCUP_V2_LOG_NAME = "test/jHiccup-2.0.7S.logV2.hlog"
# Test input and expected output values
JHICCUP_CHECKLISTS = [
    {'target': {'histogram_count': 62,
                'total_count': 48761,
                'accumulated_histogram.get_value_at_percentile(99.9)': 1745879039,
                'accumulated_histogram.get_max_value()': 1796210687,
                'log_reader.get_start_time_sec()': 1441812279.474}},
    {'range_start_time_sec': 5,
     'range_end_time_sec': 20,
     'target': {'histogram_count': 15,
                'total_count': 11664,
                'accumulated_histogram.get_value_at_percentile(99.9)': 1536163839,
                'accumulated_histogram.get_max_value()': 1544552447}},
    {'range_start_time_sec': 40,
     'range_end_time_sec': 60,
     'target': {'histogram_count': 20,
                'total_count': 15830,
                'accumulated_histogram.get_value_at_percentile(99.9)': 1779433471,
                'accumulated_histogram.get_max_value()': 1796210687}}
]

@pytest.mark.log
def test_jHiccup_v2_log():
    accumulated_histogram = HdrHistogram(LOWEST, HIGHEST, SIGNIFICANT)
    for checklist in JHICCUP_CHECKLISTS:
        accumulated_histogram.reset()
        log_reader = HistogramLogReader(JHICCUP_V2_LOG_NAME, accumulated_histogram)

        histogram_count = 0
        total_count = 0
        target_numbers = checklist.pop('target')
        while 1:
            decoded_histogram = log_reader.get_next_interval_histogram(**checklist)
            if not decoded_histogram:
                break
            histogram_count += 1
            total_count += decoded_histogram.get_total_count()
            accumulated_histogram.add(decoded_histogram)
            # These logs use 8 byte counters
            assert(decoded_histogram.get_word_size() == 8)
        for statement in target_numbers:
            assert(eval(statement) == target_numbers[statement])

        log_reader.close()

ARRAY_SIZE = 10

@pytest.mark.pyhdrh
def test_add_array_errors():
    with pytest.raises(TypeError):
        add_array()
    with pytest.raises(TypeError):
        add_array(100)
    with pytest.raises(TypeError):
        add_array(None, None, 0, 0)
    src_array = (c_uint16 * ARRAY_SIZE)()
    # negative length
    with pytest.raises(ValueError):
        add_array(addressof(src_array), addressof(src_array), -1, sizeof(c_uint16))
    # invalid word size
    with pytest.raises(ValueError):
        add_array(addressof(src_array), addressof(src_array), 0, 0)

def check_add_array(int_type):
    src_array = (int_type * ARRAY_SIZE)()
    dst_array = (int_type * ARRAY_SIZE)()
    expect_added = 0
    for index in xrange(ARRAY_SIZE):
        src_array[index] = index
        expect_added += index
    added = add_array(addressof(dst_array), addressof(src_array), ARRAY_SIZE, sizeof(int_type))
    assert(added == expect_added)
    for index in xrange(ARRAY_SIZE):
        assert(dst_array[index] == index)
    # overflow
    src_array[0] = -1
    dst_array[0] = -1
    with pytest.raises(OverflowError):
        add_array(addressof(dst_array), addressof(src_array), ARRAY_SIZE, sizeof(int_type))

@pytest.mark.pyhdrh
def test_add_array():
    for int_type in [c_uint16, c_uint32, c_uint64]:
        check_add_array(int_type)

@pytest.mark.pyhdrh
def test_zz_encode_errors():
    with pytest.raises(TypeError):
        encode()
    with pytest.raises(TypeError):
        encode(None, None, 0, 0)
    src_array = (c_uint16 * ARRAY_SIZE)()
    src_array_addr = addressof(src_array)
    dst_len = 9 * ARRAY_SIZE

    # negative length
    with pytest.raises(ValueError):
        encode(src_array_addr, -1, sizeof(c_uint16), 0, dst_len)
    # dest length too small
    with pytest.raises(ValueError):
        encode(src_array_addr, ARRAY_SIZE, 4, 0, 4)
    # invalid word size
    with pytest.raises(ValueError):
        encode(src_array_addr, ARRAY_SIZE, 3, 0, 0)
    # Null dest ptr
    with pytest.raises(ValueError):
        encode(src_array_addr, ARRAY_SIZE, 4, 0, dst_len)

def check_zz_encode(int_type):
    src_array = (int_type * ARRAY_SIZE)()
    src_array_addr = addressof(src_array)
    dst_len = 9 * ARRAY_SIZE
    dst_array = (c_uint8 * dst_len)()
    dst_array_addr = addressof(dst_array)

    res = encode(src_array_addr, ARRAY_SIZE, sizeof(int_type), dst_array_addr, dst_len)
    # should be 1 byte set to 0x13 (10 zeros => value = -10, or 0x13 in zigzag
    # encoding
    assert(res == 1)
    assert(dst_array[0] == 0x13)

    # last counter set to 1
    # the encoded result should be 2 bytes long
    # 0x11   (9 zeros => -9 coded as 17)
    # 0x02   (1 is coded as 2)
    src_array[ARRAY_SIZE - 1] = 1
    res = encode(src_array_addr, ARRAY_SIZE, sizeof(int_type), dst_array_addr, dst_len)
    assert(res == 2)
    assert(dst_array[0] == 0x11)
    assert(dst_array[1] == 0x02)

    # all counters set to 1, we should get a zigzag encoded of
    # 10 bytes all set to 0x02 (in zigzag encoding 1 is coded as 2)
    for index in xrange(ARRAY_SIZE):
        src_array[index] = 1
    res = encode(src_array_addr, ARRAY_SIZE, sizeof(int_type), dst_array_addr, dst_len)
    assert(res == ARRAY_SIZE)
    for index in xrange(ARRAY_SIZE):
        assert(dst_array[index] == 2)

@pytest.mark.pyhdrh
def test_zz_encode():
    for int_type in [c_uint16, c_uint32, c_uint64]:
        check_zz_encode(int_type)

# Few malicious V2 encodes using ZiZag LEB128/9 bytes
# Valid large value overflows smaller size dest counter
# This is the largest positive number (zigzag odd numbers are positive)
LARGE_POSITIVE_VALUE = '\xFE\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'
# This is the largest negative number
LARGE_NEGATIVE_VALUE = '\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'
#
# A simple 1 at index 0, followed by a
# large enough negative value to be dangerous: -2147483648 (smallest negative signed 32 bit)
INDEX_SKIPPER_VALUE = '\x01\x02\xFF\xFF\xFF\xFF\x0F\x02'
# Truncated end
TRUNCATED_VALUE = '\xFF\xFF'

@pytest.mark.pyhdrh
def test_zz_decode_errors():
    with pytest.raises(TypeError):
        decode(None, None, 0, 0)
    dst_array = (c_uint16 * ARRAY_SIZE)()
    # negative array size
    with pytest.raises(IndexError):
        decode(' ', 0, addressof(dst_array), -1, sizeof(c_uint16))
    # invalid word size
    with pytest.raises(ValueError):
        decode(' ', 0, addressof(dst_array), ARRAY_SIZE, 3)
    # read index negative
    with pytest.raises(IndexError):
        decode('', -1, addressof(dst_array), ARRAY_SIZE, sizeof(c_uint16))
    # Truncated end
    with pytest.raises(ValueError):
        decode(TRUNCATED_VALUE, 0, addressof(dst_array), ARRAY_SIZE, sizeof(c_uint16))
    # Too large positive value for this counter size
    with pytest.raises(OverflowError):
        decode(LARGE_POSITIVE_VALUE, 0, addressof(dst_array), ARRAY_SIZE, sizeof(c_uint16))
    # Negative overflow
    with pytest.raises(OverflowError):
        decode(LARGE_NEGATIVE_VALUE, 0, addressof(dst_array), ARRAY_SIZE, sizeof(c_uint16))
    # zero count skip index out of bounds
    with pytest.raises(IndexError):
        decode(INDEX_SKIPPER_VALUE, 0, addressof(dst_array), ARRAY_SIZE, sizeof(c_uint16))
    # read index too large => empty results
    res = decode('BUMMER', 8, addressof(dst_array), ARRAY_SIZE, sizeof(c_uint16))
    assert(res['total'] == 0)

def check_zz_identity(src_array, int_type, min_nz_index, max_nz_index, total_count, offset):
    dst_len = (sizeof(int_type) + 1) * ARRAY_SIZE
    dst = (c_uint8 * (offset + dst_len))()

    varint_len = encode(addressof(src_array), ARRAY_SIZE, sizeof(int_type),
                        addressof(dst) + offset, dst_len)
    varint_string = string_at(dst, varint_len + offset)

    dst_array = (int_type * ARRAY_SIZE)()
    res = decode(varint_string, offset, addressof(dst_array), ARRAY_SIZE, sizeof(int_type))
    assert(res['total'] == total_count)
    if total_count:
        assert(res['min_nonzero_index'] == min_nz_index)
        assert(res['max_nonzero_index'] == max_nz_index)
    for index in xrange(ARRAY_SIZE):
        assert(dst_array[index] == src_array[index])

# A large positive value that can fit 16-bit signed
ZZ_COUNTER_VALUE = 30000

def check_zz_decode(int_type, hdr_len):
    src_array = (int_type * ARRAY_SIZE)()
    check_zz_identity(src_array, int_type, 0, 0, 0, hdr_len)

    # last counter set to ZZ_COUNTER_VALUE
    # min=max=ARRAY_SIZE-1
    src_array[ARRAY_SIZE - 1] = ZZ_COUNTER_VALUE
    check_zz_identity(src_array, int_type, ARRAY_SIZE - 1,
                      ARRAY_SIZE - 1, ZZ_COUNTER_VALUE, hdr_len)

    # all counters set to ZZ_COUNTER_VALUE
    for index in xrange(ARRAY_SIZE):
        src_array[index] = ZZ_COUNTER_VALUE
    check_zz_identity(src_array, int_type, 0, ARRAY_SIZE - 1,
                      ZZ_COUNTER_VALUE * ARRAY_SIZE, hdr_len)

@pytest.mark.pyhdrh
def test_zz_decode():
    for int_type in [c_uint16, c_uint32, c_uint64]:
        for hdr_len in [0, 8]:
            check_zz_decode(int_type, hdr_len)

def hex_dump(label, str):
    print label
    print ':'.join(x.encode('hex') for x in str)
